structure IL1Context :> IL1CONTEXT =
  struct
    open IL1

    (* left is most recent *)
    type context = dec list
    type label = IL1.label

    val empty = []

    (* NB: newest dec first *)
    fun extend ctx dec = dec :: ctx
    fun extend_many ctx decs = List.revAppend (decs, ctx)

    fun list_findSome pfun l =
        case l of
            [] => NONE
          | x::xs =>
                (case pfun x of
                    NONE => list_findSome pfun xs
                  | SOME y => SOME y)

    fun lookup_expvar ctx expvar =
        list_findSome
            (fn DecExpCon (x, c) =>
                    if Variable.eq (expvar, x) then SOME c else NONE
              | _ => NONE)
            ctx

    fun lookup_convar ctx convar =
        list_findSome
            (fn DecConKind (cx, k) =>
                    if Variable.eq (convar, cx) then SOME k else NONE
              | _ => NONE)
            ctx

    fun lookup_modvar ctx modvar =
        list_findSome
            (fn DecModSig (mx, sg) =>
                    if Variable.eq (modvar, mx) then SOME sg else NONE
              | _ => NONE)
            ctx


    (*** sdecs/sbnds lookup, that makes sure the output is externally valid
         by substituting away any references to variables bound earlier ***)
    (* NB: these depend on the invariant that the labels bound by an sdecs
       or an sbnds are unique -- they search from the front *)

    (* XXX way to generalize these?  copy-paste = :( *)

  local
    fun sdecs_lookup_conlab' modv sdecs lab labs =
        case sdecs of
            [] => NONE
          | (l, DecConKind (cx, k))::sds =>
                if label_eq (l, lab) then
                    SOME k
                else
                    (*
                    sdecs_lookup_conlab modv sds lab >>= (fn k =>
                    subst_con_kind (ConModProj (modv, l)) cx k)
                    *)
                    Option.map (subst_con_kind (ConModProj (modv, l, labs)) cx)
                               (sdecs_lookup_conlab' modv sds lab labs)
          | (l, DecModSig (mx, sg))::sds =>
                Option.map (subst_mod_kind (ModModProj (modv, l, labs)) mx)
                           (sdecs_lookup_conlab' modv sds lab labs)
          | (l, DecExpCon (ex, c))::sds =>
                (* is this substitution really necessary?  phase distinction,
                   mumble mumble... *)
                Option.map (subst_exp_kind (ExModProj (modv, l, labs)) ex)
                           (sdecs_lookup_conlab' modv sds lab labs)
    in
      fun sdecs_lookup_conlab modv sdecs lab = sdecs_lookup_conlab' modv sdecs lab (labdom sdecs)
    end

  local
    fun sdecs_lookup_modlab' modv sdecs lab labs =
        case sdecs of
            [] => NONE
          | (l, DecConKind (cx, k))::sds =>
                Option.map (subst_con_signat (ConModProj (modv, l, labs)) cx)
                           (sdecs_lookup_modlab' modv sds lab labs)
          | (l, DecModSig (mx, sg))::sds =>
                if label_eq (l, lab) then
                    SOME sg
                else
                    Option.map (subst_mod_signat (ModModProj (modv, l, labs)) mx)
                               (sdecs_lookup_modlab' modv sds lab labs)
          | (l, DecExpCon (ex, c))::sds =>
                (* is this substitution really necessary?  phase distinction,
                   mumble mumble... *)
                Option.map (subst_exp_signat (ExModProj (modv, l, labs)) ex)
                           (sdecs_lookup_modlab' modv sds lab labs)
    in
      fun sdecs_lookup_modlab modv sdecs lab = sdecs_lookup_modlab' modv sdecs lab (labdom sdecs)
    end

  local
    fun sdecs_lookup_explab' modv sdecs lab labs =
        case sdecs of
            [] => NONE
          | (l, DecConKind (cx, k))::sds =>
                Option.map (subst_con_con (ConModProj (modv, l, labs)) cx)
                           (sdecs_lookup_explab' modv sds lab labs)
          | (l, DecModSig (mx, sg))::sds =>
                Option.map (subst_mod_con (ModModProj (modv, l, labs)) mx)
                           (sdecs_lookup_explab' modv sds lab labs)
          | (l, DecExpCon (ex, c))::sds =>
                if label_eq (l, lab) then
                    SOME c
                else
                    (* is this substitution really necessary?  phase
                       distinction, mumble mumble... *)
                    Option.map (subst_exp_con (ExModProj (modv, l, labs)) ex)
                               (sdecs_lookup_explab' modv sds lab labs)
    in
      fun sdecs_lookup_explab modv sdecs lab = sdecs_lookup_explab' modv sdecs lab (labdom sdecs)
    end


  local
    fun sbnds_lookup_conlab' modv sbnds lab labs =
        case sbnds of
            [] => NONE
          | (l, BndCon (cx, c))::sbs =>
                if label_eq (l, lab) then
                    SOME c
                else
                    Option.map (subst_con_con (ConModProj (modv, l, labs)) cx)
                               (sbnds_lookup_conlab' modv sbs lab labs)
          | (l, BndMod (mx, m))::sbs =>
                Option.map (subst_mod_con (ModModProj (modv, l, labs)) mx)
                           (sbnds_lookup_conlab' modv sbs lab labs)
          | (l, BndExp (ex, e))::sbs =>
                (* is this substitution really necessary?  phase distinction,
                   mumble mumble... *)
                Option.map (subst_exp_con (ExModProj (modv, l, labs)) ex)
                           (sbnds_lookup_conlab' modv sbs lab labs)
    in
      fun sbnds_lookup_conlab modv sdecs lab = sbnds_lookup_conlab' modv sdecs lab (labdom sdecs)
    end

  local
    fun sbnds_lookup_modlab' modv sbnds lab labs =
        case sbnds of
            [] => NONE
          | (l, BndCon (cx, c))::sbs =>
                Option.map (subst_con_module (ConModProj (modv, l, labs)) cx)
                           (sbnds_lookup_modlab' modv sbs lab labs)
          | (l, BndMod (mx, m))::sbs =>
                if label_eq (l, lab) then
                    SOME m
                else
                    Option.map (subst_mod_module (ModModProj (modv, l, labs)) mx)
                               (sbnds_lookup_modlab' modv sbs lab labs)
          | (l, BndExp (ex, e))::sbs =>
                (* is this substitution really necessary?  phase distinction,
                   mumble mumble... *)
                Option.map (subst_exp_module (ExModProj (modv, l, labs)) ex)
                           (sbnds_lookup_modlab' modv sbs lab labs)
    in
      fun sbnds_lookup_modlab modv sdecs lab = sbnds_lookup_modlab' modv sdecs lab (labdom sdecs)
    end


    (* NB: careful, this doesn't reverse the decs -- it expects them in
       "most-recent first" order.  (this is used primarily by
       TransContext.erase_labels, which uses the same order.) *)
    fun from_decs decs = decs
    fun to_decs decs = decs
  end
