signature IL2 =
  sig
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var

    datatype kind =
        KndType
      | KndPi of convar * kind * kind (* convar bound in kind2 *)
      | KndUnit
      | KndSigma of convar * kind * kind (* convar bound in kind2 *)
      | KndSing of con

    and con =
        ConVar of convar
      | ConAbs of convar * kind * con
      | ConApp of con * con
      | ConUnit
      | ConPair of con * con
      | ConProj1 of con
      | ConProj2 of con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)

      | TyLit of Literal.lit_ty
      | TyArrow of con * con
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of con list
      | TyProd of con list
      | TyForall of convar * kind * con (* convar bound in con *)

    datatype exp =
        ExVar of expvar
      | ExLit of Literal.lit
      | ExLam of expvar * con * exp
      | ExUnOp of Literal.unop * exp
      | ExBinOp of exp * Literal.binop * exp
      | ExFix of (expvar * expvar * con * con * exp) list
      | ExApp of exp * exp
      | ExHandle of exp * exp
      | ExRaise of con * exp
      | ExRef of exp
      | ExGet of exp
      | ExSet of exp * exp
      | ExRoll of con * exp
      | ExUnroll of exp
      | ExProd of exp list
      | ExProj of int * exp
      | ExSumInj of con * int * exp
      | ExCase of con * exp * exp list
      | ExTagInj of exp * exp
      | ExNewTag of con
      | ExIfTagOf of exp * exp * exp * exp
      | ExConAbs of convar * kind * exp (* convar bound in exp *)
      | ExConApp of exp * con
      | ExLet of expvar * exp * exp (* expvar bound in exp2 *)

    (* degenerate module calculus *)
    and module = Mod of con * exp
    and signat = Sig of convar * kind * con (* convar bound in con *)

  (* we need to make this a separate type since mlton doesn't
    support withtype in signatures *)
    type fbnd = (expvar * expvar * con * con * exp)


    (*** derived forms ***)
    val labelledKndSing : con -> kind -> kind

    val kndPair : kind * kind -> kind
    val kndProd : kind list -> kind

    val conProd : con list -> con
    val conProj : int * con -> con

    val tyUnit : con
    val tyInt : con
    val tyBool : con
    val tyChar : con
    val tyString : con

    val exUnit : exp
    val exBool : bool -> exp
    val exInt : int -> exp
    val exChar : char -> exp
    val exString : string -> exp


    (*** mapping ***)

    val map_kind : (convar * kind -> convar * kind)
                -> (con -> con)
                -> kind -> kind
    val map_con : (convar -> con)
               -> (convar * con -> convar * con)
               -> (kind -> kind)
               -> con -> con
    val map_exp : (expvar -> exp)
               -> (fbnd list -> fbnd list)
               -> (expvar * exp -> expvar * exp)
               -> (convar * exp -> convar * exp)
               -> (con -> con)
               -> (kind -> kind)
               -> exp -> exp


    (* alpha variation *)

    (* [alpha_vary (var, phrase) replacefun] yields (var', phrase'), where
       var' is fresh, and (var. phrase) is alpha-equal to (var'. phrase') *)
    val alpha_vary : 'a Variable.var * 'b
                  -> ('a Variable.var -> 'a Variable.var -> 'b -> 'b)
                  -> 'a Variable.var * 'b
    val alpha_vary2 : 'a Variable.var * 'b
                   -> 'a Variable.var * 'b
                   -> ('a Variable.var -> 'a Variable.var -> 'b -> 'c)
                   -> 'a Variable.var * ('c * 'c)


    (*** replacement of convars ***)

    val replace_con_kind : convar -> convar -> kind -> kind
    val replace_con_con : convar -> convar -> con -> con
    val replace_con_exp : convar -> convar -> exp -> exp
    val replace_exp_exp : expvar -> expvar -> exp -> exp


    (*** substitution for convars ***)

    val subst_con_kind : con -> convar -> kind -> kind
    val subst_con_con : con -> convar -> con -> con
    val subst_con_exp : con -> convar -> exp -> exp

    val subst_con_cons : con -> convar -> con list -> con list
  end
